use strict;
use warnings;
use Math::Trig qw( pi );

srand( 2 ); # 好みの音が得られるように初期化
my @noise = map { rand( 2.0 ) - 1.0; } 1..1024;

my %func_table = (
    'pulse' => sub { # 矩形波
        return ( $_[0] < 0.5 ) ? -1.0 : 1.0;
    },
    'sin'   => sub { # サイン波（正弦波）
        return sin( 2.0 * pi() * $_[0] );
    },
    'saw'   => sub { # のこぎり波
        return ( 2.0 * $_[0] ) - 1.0;
    },
    'tri'   => sub { # 三角波
        if ( $_[0] < 0.5 ) {
            # -1.0 -> +1.0
            return -1.0 + ( 4.0 * $_[0] );
        }
        else {
            # +1.0 -> -1.0
            return 1.0 - ( 4.0 * ($_[0] - 0.5) );
        }
    },
    'noise' => sub { # ノイズ
        if ( $_[0] < 1.0 ) {
            my $idx = int( $_[0] * scalar(@noise) );
            return $noise[$idx];
        }
        else {
            return 0.0;
        }
    }
);

# 種類を指定して、波形を生成する関数を返す
sub create_mod_func {
    my $func = $func_table{$_[0]} or die;
    return $func;
}

# 任意のパラメータで波形を生成する関数を返す
sub create_modulator {
    my $samples_per_sec = shift;
    my $arg_ref = shift;

    my $freq = $arg_ref->{freq};
    my $osc_func = create_mod_func( $arg_ref->{waveform} );
    my $t = 0.0;
    my $samples_per_cycle = $samples_per_sec / $freq;
    return sub {
        my $mod = shift;
        my $ret = $osc_func->(
            $t / $samples_per_cycle
        );

        my $dt = 1.0 + $mod;
        if ( 0.0 < $dt ) {
            $t += $dt;
            while ( $samples_per_cycle <= $t ) {
                $t -= $samples_per_cycle;
            }
        }

        return $ret;
    };
}

# 変調用の波形を生成する関数を返す
sub create_pitch_modulator {
    my $samples_per_sec = shift;
    my $arg_ref = shift;

    my $waveform = $arg_ref->{waveform};
    my $depth = $arg_ref->{depth};
    if ( $waveform eq 'env' ) {
        my $curve = 1.0;
        if ( exists $arg_ref->{curve} ) {
            $curve = $arg_ref->{curve};
        }
        my $env = create_envelope(
            $samples_per_sec,
            { sec => $arg_ref->{speed}, curve => $curve } );
        return sub { $env->() * $depth };
    }
    else {
        my $osc = create_modulator(
            $samples_per_sec,
            {
                freq => (1.0 / $arg_ref->{speed}),
                waveform => $waveform
            }
        );
        return sub { $osc->(0) * $depth };
    }
}

# エンベロープ（減衰カーブ）を生成する関数を返す
sub create_envelope {
    my $samples_per_sec = shift;
    my $arg_ref = shift;

    my $curve = 1.0;
    if ( exists $arg_ref->{curve} ) {
        $curve = $arg_ref->{curve};
    }
    my $mod_func = sub { return ( 1.0 - $_[0] ); };
    my $t = 0.0;
    my $interval = $samples_per_sec * $arg_ref->{sec};
    return sub {
        if ( $t < $interval ) {
            my $ret = $mod_func->( $t / $interval );
            $t += 1.0;
            return $ret ** $curve;
        }
        else {
            return 0.0;
        }
    };
}

# 単音の波形データの生成する
sub create_oneshot {
    my $samples_per_sec = 44100;
    my $arg_ref = shift;

    my $osc = create_modulator(
        $samples_per_sec, $arg_ref->{osc} );
    my $env = create_envelope(
        $samples_per_sec, $arg_ref->{amp} );

    my $mod = sub { return 0.0; }; # pitch modulator
    if ( exists $arg_ref->{osc}->{mod} ) {
        $mod = create_pitch_modulator(
            $samples_per_sec, $arg_ref->{osc}->{mod} );
    }

    my $amp = $arg_ref->{amp};
    my $gate_time = int( $samples_per_sec * $amp->{sec} );

    my @samples = map {
        $osc->( $mod->() ) * $env->();
    } 0..($gate_time - 1);

    return \@samples;
}

# チャンネルデータの生成する
sub create_channel {
    my $samples_per_sec = 44100;
    my $bpm = shift; # beats per minute
    my $arg_ref = shift;

    my $seq = $arg_ref->{seq};
    my $tone = $arg_ref->{tone};

    # 単音の波形データの生成
    my $oneshot_ref = create_oneshot( $tone );

    # 発音タイミングの計算
    my $bps = ( $bpm / 60.0 ) * 4;
    my $seq_cnt = scalar( @{$seq} );
    my $interval = $samples_per_sec / $bps;
    my @plot_tmpl = ();
    for (my $i=0; $i<$seq_cnt; $i++) {
        push @plot_tmpl, int($i * $interval);
    }

    # 必要な配列サイズを計算して初期化
    my $wav_size =
        $plot_tmpl[$seq_cnt - 1] + scalar(@{$oneshot_ref});
    my @channel = map { 0.0; } 1..$wav_size;

    # チャンネルデータの生成
    for (my $i=0; $i<$seq_cnt; $i++) {
        if ( not $seq->[$i] ) { next; }
        my $j = $plot_tmpl[$i];
        map { $channel[$j++] += $_; } @{$oneshot_ref};
    }

    return \@channel;
}

# WAVEファイル出力
sub save_as_wav {
    my $samples_per_sec = 44100;
    my $bits_per_sample = 16;
    my $file_name = shift;
    my $samples_ref = shift;

    my $block_size = $bits_per_sample / 8;
    my $size = scalar(@{$samples_ref}) * $block_size;
    my $bytes_per_sec = $block_size * $samples_per_sec;
    my $header =
          'RIFF'                        # ChunkID
        . pack('L', ($size + 36))       # ChunkSize
        . 'WAVE';                       # FormType
    my $fmt_chunk =
          'fmt '                        # ChunkID
        . pack('L', 16)                 # ChunkSize
        . pack('S', 1)                  # WaveFormatType
        . pack('S', 1)                  # Channel
        . pack('L', $samples_per_sec)   # SamplesPerSec
        . pack('L', $bytes_per_sec)     # BytesPerSec
        . pack('S', $block_size)        # BlockSize
        . pack('S', $bits_per_sample);  # BitsPerSample
    my $data_chunk =
          'data'                        # ChunkID
        . pack('L', $size);             # ChunkSize

    open( my $fh, '>', $file_name ) or die;
    binmode $fh;
    print $fh ($header . $fmt_chunk . $data_chunk);
    foreach my $sample (@{$samples_ref}) {
        print $fh pack( 's', int($sample * 32767.0) );
    }
    close $fh;
}

# 440Hzのサイン波を生成する例
my $osc = create_modulator(
    44100,                  # サンプリング周波数
    {
        freq => 440,        # 周波数
        waveform => 'sin'   # 波形の種類
    }
);
my @sin_samples = map {
    $osc->( 0 );
} 1..44100;

# 周波数変調の例
my $mod = create_pitch_modulator(
    44100,                  # サンプリング周波数
    {
        speed => 0.5,       # 変調する速度
        depth => 0.5,       # 変調の度合い
        waveform => 'pulse' # 矩形波
    }
);
my @sin_with_mod_samples = map {
    $osc->( $mod->() );
} 1..44100;

# キックの音色の例
my $kick = {
    osc => {
        freq => 25,
        waveform => 'sin',
        mod => {
            speed => 0.25, depth => 3.5, waveform => 'env', curve => 1.8
        }
    },
    amp => {
        sec => 0.25, waveform => 'env', curve => 1.4
    }
};

# スネアの音色の例
my $snare = {
    osc => {
        freq => 400,
        waveform => 'tri',
        mod => {
            speed => 0.08, depth => 9.0, waveform => 'noise'
        }
    },
    amp => { sec => 0.16, waveform => 'env', curve => 2.2 }
};

# オープンハイハットの音色の例
my $o_hat = {
    osc => {
        freq => 16000,
        waveform => 'tri',
        mod => {
            speed => 0.06, depth => 6.0, waveform => 'noise'
        }
    },
    amp => { sec => 0.15, waveform => 'env', curve => 2.7 }
};

# クローズハイハットの音色の例
my $c_hat = {
    osc => {
        freq => 16000,
        waveform => 'tri',
        mod => {
            speed => 0.06, depth => 6.0, waveform => 'noise'
        }
    },
    amp => { sec => 0.08, waveform => 'env', curve => 2.7 }
};

# 単音の波形データを生成する例
my $samples_kick_ref  = create_oneshot( $kick  );
my $samples_snare_ref = create_oneshot( $snare );
my $samples_o_hat_ref = create_oneshot( $o_hat );
my $samples_h_hat_ref = create_oneshot( $c_hat );

# beatsのパラメータ
my $bpm = 138; # beats per minute
my $seq_kick  = [ 1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0 ];
my $seq_snare = [ 0,0,0,0,1,0,0,0,0,0,0,0,1,0,0,0 ];
my $seq_o_hat = [ 0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0 ];
my $seq_c_hat = [ 1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1 ];
my @beats = (
    { seq => $seq_kick,  tone => $kick,  vol => 1.00 },
    { seq => $seq_snare, tone => $snare, vol => 0.12 },
    { seq => $seq_o_hat, tone => $o_hat, vol => 0.06 },
    { seq => $seq_c_hat, tone => $c_hat, vol => 0.02 }
);

# チャンネルデータの生成
my @channels = map {
    create_channel( $bpm, $_ );
} @beats;

# ミックス
my @samples_beats = ();
foreach my $ch_info ( @beats ) {
    my $ch = create_channel( $bpm, $ch_info );
    my $vol = $ch_info->{vol};
    for (my $i=0; $i<scalar(@{$ch}); $i++) {
        $samples_beats[$i] += ( $ch->[$i] * $vol );
    }
}

# クリップ
@samples_beats = map {
    ( 1.0 < $_ ) ? 1.0 : ( ($_ < -1.0) ? -1.0 : $_ );
} @samples_beats;

# WAVEファイル出力の例
save_as_wav( 'sin.wav', \@sin_samples );
save_as_wav( 'sin_with_mod.wav', \@sin_with_mod_samples );
save_as_wav( 'kick.wav', $samples_kick_ref );
save_as_wav( 'snare.wav', $samples_snare_ref );
save_as_wav( 'o_hat.wav', $samples_o_hat_ref );
save_as_wav( 'c_hat.wav', $samples_h_hat_ref );
save_as_wav( 'samples_beats.wav', \@samples_beats );

__END__
